/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

// the code for verification of the ECC curve params used by MS Play Ready
// libstd_cai_client_drm_msplayready.so binary library

import java.lang.*;
import java.math.*;

public class Test {
 //modular bits cnt
 static int MOD_BITS = 0x100;

 //target platform word bits cnt
 static int WORD_BITS = 0x20;

 //prime
 static String P="ffffffff00000001000000000000000000000000ffffffffffffffffffffffff";
 static BigInteger p_bi;

 //param a
 static String A="fffffffc00000004000000000000000000000003fffffffffffffffffffffffc";
 static BigInteger a_bi;

 //param b
 static String B="dc30061d04874834e5a220abf7212ed6acf005cd78843090d89cdf6229c4bddf";
 static BigInteger b_bi;

 //generator.X
 static String GX="18905f76a53755c679fb732b7762251075ba95fc5fedb60179e730d418a9143c";
 static BigInteger gx_bi;

 //generator.Y
 static String GY="8571ff1825885d85d2e88688dd21f3258b4ab8e4ba19e45cddf25357ce95560a";
 static BigInteger gy_bi;

 //2^32 factor value
 static BigInteger modfactor32_bi;
 static BigInteger inv_modfactor32_bi;

 //2^256 factor value
 static BigInteger modfactor256_bi;
 static BigInteger inv_modfactor256_bi;

 static {
  p_bi=new BigInteger(P,16);
  a_bi=new BigInteger(A,16);
  b_bi=new BigInteger(B,16);

  gx_bi=new BigInteger(GX,16);
  gy_bi=new BigInteger(GY,16);

  BigInteger two_bi=new BigInteger("2");
  modfactor32_bi=two_bi.pow(WORD_BITS);

  modfactor256_bi=two_bi.pow(MOD_BITS).mod(p_bi);

  BigInteger m1_bi=new BigInteger("-1");
  inv_modfactor32_bi=modfactor32_bi.modPow(m1_bi,p_bi);

  inv_modfactor256_bi=modfactor256_bi.modPow(m1_bi,p_bi);
 }

 private static void print(String s,BigInteger bi) {
  System.out.println(s+": "+bi.toString(16));
 }

 private static BigInteger make_bi(String s) {
  return new BigInteger(s,16);
 }

 private static BigInteger to_modular(BigInteger input) {
  int cnt=MOD_BITS/WORD_BITS;

  BigInteger r_bi=input;

  for(int i=0;i<cnt;i++) {
   r_bi=r_bi.multiply(modfactor32_bi);

   r_bi=r_bi.mod(p_bi);
  }  

  return r_bi;
 }

 private static BigInteger from_modular(BigInteger input) {
  int cnt=MOD_BITS/WORD_BITS;

  BigInteger r_bi=input;

  for(int i=0;i<cnt;i++) {
   r_bi=r_bi.multiply(inv_modfactor32_bi);

   r_bi=r_bi.mod(p_bi);
  }  

  return r_bi;
 }

 private static BigInteger to_point(BigInteger input) {
  BigInteger res=input.multiply(modfactor256_bi).mod(p_bi);

  return res;
 }

 private static BigInteger from_point(BigInteger input) {
  BigInteger res=input.multiply(inv_modfactor256_bi).mod(p_bi);

  return res;
 }

 private static BigInteger power2(BigInteger input) {
  BigInteger res=input.multiply(input);

  return res;
 }

 private static boolean ecaffine_on_curve(BigInteger xp_bi,BigInteger yp_bi) {
  BigInteger x_pow_2=from_point(power2(xp_bi));
  print("x_pow_2",x_pow_2);

  BigInteger x_pow_2_plus_a=x_pow_2.add(a_bi).mod(p_bi);
  print("x_pow_2_plus_a",x_pow_2_plus_a);

  BigInteger x_pow_3_plus_ax=from_point(x_pow_2_plus_a.multiply(xp_bi));
  print("x_pow_3_plus_ax",x_pow_3_plus_ax);

  BigInteger x_pow_3_plus_ax_plus_b=x_pow_3_plus_ax.add(b_bi).mod(p_bi);
  print("x_pow_3_plus_ax_plus_b",x_pow_3_plus_ax_plus_b);

  BigInteger y_pow_2=from_point(power2(yp_bi));
  print("y_pow_2",y_pow_2);

  return x_pow_3_plus_ax_plus_b.equals(y_pow_2);
 }

 public static void main(String args[]) throws Throwable {
  try {
   System.out.println("EC test");

   print("p",p_bi);
   print("a",a_bi);
   print("b",b_bi);

   String X="c8b6af16ee941aadaa5389b4af2c10e356be42af175ef3face93254e7b0b3d9b";
   BigInteger x_bi=make_bi(X);

   print("x",x_bi);

   String Y="982b27b5cb2341326e56aa857dbfd5c634ce2cf9ea74fca8f2af5957efeea562";

   BigInteger y_bi=make_bi(Y);

   print("y",y_bi);
   System.out.println("");

   BigInteger xp_bi=to_point(x_bi);
   print("X to_point",xp_bi);

   BigInteger yp_bi=to_point(y_bi);
   print("Y to_point",yp_bi);

   System.out.println("");

   boolean res=ecaffine_on_curve(xp_bi,yp_bi);
   System.out.println("ecaffine_on_curve: "+res);

   System.out.println("");

   BigInteger real_a_bi=from_point(a_bi);
   print("real_a_bi",real_a_bi);

   BigInteger real_b_bi=from_point(b_bi);
   print("real_b_bi",real_b_bi);

   System.out.println("");
 
   BigInteger real_genx_bi=from_point(gx_bi);
   print("real_genx_bi",real_genx_bi);
   
   BigInteger real_geny_bi=from_point(gy_bi);
   print("real_geny_bi",real_geny_bi);
  } catch(Throwable e) {
   e.printStackTrace();
  }
 }
} 
