/*## (c) SECURITY EXPLORATIONS    2016 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2020 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package agsecres.tool;

import java.lang.*;
import java.lang.reflect.*;
import java.util.*;
import java.io.*;

public class ModRegistry {
 public static String MOD_PATH  = "mod";
 public static String MOD_CLASS = "Init";

 static Vector<TolkitMod> registry;
 static CmdDesc cmd_table[];

 static {
  registry=new Vector<TolkitMod>();
  cmd_table=new CmdDesc[0];
 }

 static boolean has_cmd(CmdDesc[] tab,String name) {
  for(int i=0;i<tab.length;i++) {
   CmdDesc cmd=tab[i];

   if (cmd.name.equals(name)) return true;
  }

  return false;
 }

 public static int size() {
  return registry.size();
 }

 public static TolkitMod get(int i) {
  if (i<size()) return registry.elementAt(i);

  return null;
 }

 public static TolkitMod get(String name) {
  for(int i=0;i<size();i++) {
   TolkitMod mod=get(i);

   if (mod.name().equals(name)) return mod;
  }

  return null;
 }

 public static boolean is_loaded(String name) {
  return get(name)!=null;
 }


 static CmdDesc[] merge_tables(CmdDesc[] mod_table,CmdDesc[] base_tab) {
  Vector<CmdDesc> vcmds=new Vector<CmdDesc>();

  for(int i=0;i<base_tab.length;i++) {
   CmdDesc cmd=base_tab[i];

   if (!has_cmd(mod_table,cmd.name)) vcmds.add(cmd);
  }

  for(int i=0;i<mod_table.length;i++) {
   CmdDesc cmd=mod_table[i];

   vcmds.add(cmd);
  }

  int num=vcmds.size();

  CmdDesc cmd_table[]=new CmdDesc[num];

  for(int i=0;i<num;i++) {
   cmd_table[i]=vcmds.elementAt(i);
  }

  return cmd_table;
 }

 static void update_tables() {
  CmdDesc res_cmd_table[]=new CmdDesc[0];

  for(int i=0;i<size();i++) {
   TolkitMod mod=get(i);

   CmdDesc mod_cmd_table[]=mod.cmd_table();
   
   res_cmd_table=merge_tables(mod_cmd_table,res_cmd_table);
  }

  cmd_table=res_cmd_table;
 }

 public static boolean load_module(String name) {
  boolean res=false;

  if (is_loaded(name)) return res;

  try {
   Class c=Class.forName(MOD_PATH+"."+name+"."+MOD_CLASS);

   TolkitMod mod=(TolkitMod)c.newInstance();
   registry.add(mod);   

   update_tables();
   res=true;
  } catch(Throwable t) {}

  return res;
 }

 public static CmdDesc[] cmd_table() {
  return cmd_table;
 }

}
