/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package agsecres.helper;

import java.lang.*;
import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.security.*;

public class JSON {
 public static abstract class JSONObject {
  public void print(int lvl,PrintWriter out) {
  }

  public int size() {
   return 0;
  }

  public JSONObject get(String name) {
   return null;
  }

  public JSONObject get(int idx) {
   return null;
  }

 }

 public static class JSONString extends JSONObject {
  String val;

  public JSONString(String s) {
   val=s;
  }

  public String val() {
   return val;
  }

  public void print(int lvl,PrintWriter out) {
   String line="";
   line=dopad(lvl*2)+"  "+val;
   out.println(line);
  }

 }

 public static class JSONBucket extends JSONObject {
  Hashtable lookup;
  Vector props;

  public JSONBucket() {
   lookup=new Hashtable();
   props=new Vector();
  }

  public void put(String prop,JSONObject val) {
   lookup.put(prop,val);
   props.add(prop);   
  }

  public void print(int lvl,PrintWriter out) {
   String line="";

   line=dopad(lvl*2)+"{";
   out.println(line);

   for(int i=0;i<props.size();i++) {
    String prop=(String)props.elementAt(i);
    JSONObject val=(JSONObject)lookup.get(prop);

    if (val instanceof JSONString) {
     JSONString str=(JSONString)val;
     line=dopad(lvl*2)+"  "+pad(prop,PAD)+" : "+str.val;
     out.println(line);
    } else {
     line=dopad(lvl*2)+"  "+pad(prop,PAD)+" : ";
     out.println(line);
     val.print(lvl+1,out);
    }
   }

   line=dopad(lvl*2)+"}";
   out.println(line);
  }

  public int size() {
   return props.size();
  }

  public JSONObject get(String name) {
   return (JSONObject)lookup.get(name);
  }

  public JSONObject get(int idx) {
   String prop=(String)props.elementAt(idx);

   if (prop!=null) {
    return (JSONObject)lookup.get(prop);
   } else return null;
  }
 }

 public static class JSONArray extends JSONObject {
  Vector elements;

  public JSONArray() {
   elements=new Vector();
  }

  public void add(JSONObject elem) {
   elements.add(elem);
  }

  public void print(int lvl,PrintWriter out) {
   String line="";

   line=dopad(lvl*2)+"[";
   out.println(line);

   for(int i=0;i<elements.size();i++) {
    JSONObject elem=(JSONObject)elements.elementAt(i);
    elem.print(lvl+1,out);
   }

   line=dopad(lvl*2)+"]";
   out.println(line);
  }

  public int size() {
   return elements.size();
  }

  public JSONObject get(int idx) {
   return (JSONObject)elements.elementAt(idx);
  }
 }

 public static class Parser {
  public static String parse_string(PushbackInputStream pbis) {
   String res=null;
   
   try {
    ByteArrayOutputStream baos=new ByteArrayOutputStream();

    int ch=-1;
    int state=0;
    boolean done=false;

    while((!done)&&((ch=pbis.read())!=-1)) {
     switch (state) {
      case 0:
       if (ch=='"') state=1;
       break;
      case 1:
       if (ch=='"') {
        done=true;
       } else baos.write(ch);
       break;
     }     
    }

    res=new String(baos.toByteArray());
   } catch(Throwable t) {
    t.printStackTrace();
   } 

   return res;
  }

  public static String parse_unknown(PushbackInputStream pbis) {
   String res=null;
   
   try {
    ByteArrayOutputStream baos=new ByteArrayOutputStream();

    int ch=-1;
    int state=0;

    while((ch=pbis.read())!=-1) {
     if (Character.isWhitespace(ch)) break;
      else
     if (!Character.isLetter(ch)) {
      pbis.unread(ch);
      break;
     } else baos.write(ch);
    }

    res=new String(baos.toByteArray());
   } catch(Throwable t) {
    t.printStackTrace();
   } 

   return res;
  }

  public static String parse_number(PushbackInputStream pbis) {
   String res=null;

   try {
    ByteArrayOutputStream baos=new ByteArrayOutputStream();

    int ch=-1;

    while((ch=pbis.read())!=-1) {
     if (Character.isDigit(ch)) {
      baos.write(ch);
     } else {
      pbis.unread(ch);
      break;
     }
    }

    res=new String(baos.toByteArray());
   } catch(Throwable t) {
    t.printStackTrace();
   } 

   return res;
  }

  public static JSONObject parse_value(PushbackInputStream pbis) {
   JSONObject obj=null;

   try {
    int ch=-1;
    boolean done=false;

    while((!done)&&((ch=pbis.read())!=-1)) {
     if (Character.isWhitespace(ch)) continue;

     if (ch=='"') {
      pbis.unread(ch);
      String s=parse_string(pbis);
      obj=new JSONString(s);
      break;
     } else

     if (Character.isDigit(ch)) {
      pbis.unread(ch);
      String s=parse_number(pbis);
      obj=new JSONString(s);
      break;
     } else

     if (ch=='{') {
      obj=parse_object(pbis);
      break;
     } else

     if (ch=='[') {
      obj=parse_array(pbis);
      break;
     } else {
      pbis.unread(ch);
      String s=parse_unknown(pbis);
      obj=new JSONString(s);
      break;
     }
    }
   } catch(Throwable t) {
    t.printStackTrace();
   } 

   return obj;
  }

  public static JSONObject parse_array(PushbackInputStream pbis) {
   JSONArray obj=null;

   try {
    int ch=-1;
    boolean done=false;

    while((!done)&&((ch=pbis.read())!=-1)) {
     if (Character.isWhitespace(ch)) continue;

     if (ch=='"') {
      pbis.unread(ch);
      String s=parse_string(pbis);
      JSONObject elem=new JSONString(s);
      if (obj==null) obj=new JSONArray();
      obj.add(elem);
     } else

     if (ch==',') continue;
      else

     if (ch=='{') {
      JSONObject elem=parse_object(pbis);
      if (obj==null) obj=new JSONArray();
      obj.add(elem);
     } else

     if (ch==']') {
      done=true;
      break;
     }
    }
    
   } catch(Throwable t) {
    t.printStackTrace();
   } 

   return obj;
  }

  public static JSONObject parse_object(PushbackInputStream pbis) {
   JSONBucket obj=null;

//   System.out.println("--- PARSE OBJECT");   
   try {
    String prop=null;
    JSONObject val=null;

    int state=0;
    int ch=-1;
    boolean done=false;

    while((!done)&&((ch=pbis.read())!=-1)) {
     if (Character.isWhitespace(ch)) continue;

     if (ch==',') continue;
      else

     if (ch=='}') {
      done=true;
      break;
     }

     if (state==0) {
      if (ch=='"') {
       pbis.unread(ch);
       prop=parse_string(pbis);
//System.out.println("prop: "+prop);
      } else

      if (ch==':') {
       state=1;
      }
     } else {
      pbis.unread(ch);
      val=parse_value(pbis);

//      if (val instanceof JSONString) {
//       System.out.println("val: "+((JSONString)val).val);
//      }
     }

     if ((prop!=null)&&(val!=null)) {
      if (obj==null) obj=new JSONBucket();
      obj.put(prop,val);

      prop=null;
      val=null;
      state=0;
     }
    }

   } catch(Throwable t) {
    t.printStackTrace();
   } 

   if (obj==null) obj=new JSONBucket();

   return obj;
  }

  public static JSONObject parse(InputStream is) {
   JSONObject res=null;

   try {
    PushbackInputStream pbis=new PushbackInputStream(is);

    int ch=-1;

    while((ch=pbis.read())!=-1) {
     if (ch=='{') {
      res=parse_object(pbis);
     } else break;
    }

   } catch(Throwable t) {
    t.printStackTrace();
   } 

   return res;
  }
 }

 public static final int PAD = 20;

 static String dopad(int num) {
  String pad="";

  for(int i=0;i<num;i++) {
   pad+=" ";
  }

  return pad;
 }

 static String pad(String s,int num) {
  String pad=s;

  int cnt=num-s.length();

  for(int i=0;i<cnt;i++) {
   pad+=" ";
  }

  return pad;
 }

 public static String get_string(JSONObject root,String name) {
  String res=null;

  try {
   if (root instanceof JSONBucket) {
    JSONBucket bucket=(JSONBucket)root;

    JSONObject type=bucket.get(name);
    if (type instanceof JSONString) {
     res=((JSONString)type).val();
    }
   }
  } catch(Throwable t) {}

  return res;
 }

 public static JSONBucket get_bucket(JSONObject root,String name) {
  JSONBucket res=null;

  try {
   if (root instanceof JSONBucket) {
    JSONBucket bucket=(JSONBucket)root;

    JSONObject type=bucket.get(name);
    if (type instanceof JSONBucket) {
     res=(JSONBucket)type;
    }
   }
  } catch(Throwable t) {}

  return res;
 }

 public static JSONArray get_array(JSONObject root,String name) {
  JSONArray res=null;

  try {
   if (root instanceof JSONBucket) {
    JSONBucket bucket=(JSONBucket)root;

    JSONObject type=bucket.get(name);
    if (type instanceof JSONArray) {
     res=(JSONArray)type;
    }
   }
  } catch(Throwable t) {}

  return res;
 }

 public static JSONObject parse(String xml) throws Throwable {
  return parse(xml.getBytes());
 }

 public static JSONObject parse(byte data[]) throws Throwable {
  JSONObject res=null;

  try {
   ByteArrayInputStream bais=new ByteArrayInputStream(data);

   res=Parser.parse(bais);
  } catch(Throwable t) {
   t.printStackTrace();
  } 

  return res;
 }

 public static void fill_instance(Object instance,Class clazz,JSONObject obj) {
  try {
   Field ftab[]=clazz.getDeclaredFields();

   for(int i=0;i<ftab.length;i++) {
    Field f=ftab[i];

    String fname=f.getName();

    if (fname.startsWith("att")) {
     fname=fname.substring(3);

     String value=JSON.get_string(obj,fname);

     if (value!=null) {
      f.setAccessible(true);
      f.set(instance,value);
     }    
    }
   }
  } catch(Throwable t) {
   t.printStackTrace();
  }
 }

 public static Object instance_from_node(Class clazz,JSONObject obj) {
  Object instance=null;

  try {
   instance=clazz.newInstance();

   while(clazz!=null) {
    fill_instance(instance,clazz,obj);

    clazz=clazz.getSuperclass();
   }
  } catch(Throwable t) {
   instance=null;
   t.printStackTrace();
  }
   
  return instance;
 }

}
