/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

//basic implementation of RSA crypto

package agsecres.helper;

import agsecres.tool.*;
import java.lang.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.nio.*;
import java.util.zip.*;
import javax.crypto.*;
import java.math.BigInteger;
import java.security.*;
import java.security.interfaces.*;
import java.security.spec.*;
import javax.crypto.spec.*;
import javax.crypto.*;

public class RSA {
 public static class KeyPair {
  RSAPublicKeySpec pubKeySpec;
  RSAPrivateKeySpec privKeySpec;

  RSAPublicKey pubKey;
  RSAPrivateKey privKey;
  int keysize;

  public void init_keypair(byte prvkey[],byte pubkey[]) {
   try {
    byte prv_mod[]=new byte[keysize];
    byte prv_exp[]=new byte[keysize];

    System.arraycopy(prvkey,0,prv_mod,0,keysize);
    System.arraycopy(prvkey,keysize,prv_exp,0,keysize);

    byte pub_mod[]=new byte[keysize];
    byte pub_exp[]=new byte[keysize];

    System.arraycopy(pubkey,0,pub_mod,0,keysize);
    System.arraycopy(pubkey,keysize,pub_exp,0,keysize);

    String PRIV_MOD=Utils.construct_hex_string(prv_mod);
    String PRIV_EXP=Utils.construct_hex_string(prv_exp);

    String PUB_MOD=Utils.construct_hex_string(pub_mod);
    String PUB_EXP=Utils.construct_hex_string(pub_exp);

    pubKeySpec=new RSAPublicKeySpec(
                      new BigInteger(PUB_MOD,16),
                      new BigInteger(PUB_EXP,16));
    privKeySpec=new RSAPrivateKeySpec(
                      new BigInteger(PRIV_MOD,16),  
                      new BigInteger(PRIV_EXP,16));
        
    KeyFactory keyFactory=KeyFactory.getInstance("RSA");

    pubKey=(RSAPublicKey)keyFactory.generatePublic(pubKeySpec);
    privKey=(RSAPrivateKey)keyFactory.generatePrivate(privKeySpec);
   } catch(Throwable t) {
    t.printStackTrace();
   }
  }

  public KeyPair(byte prvkey[],byte pubkey[]) {
   keysize=prvkey.length>>1;
 
   init_keypair(prvkey,pubkey);
  }

  public RSAPublicKey pub() {
   return pubKey;
  }

  public RSAPrivateKey prv() {
   return privKey;
  }

  public BigInteger prv_mod() {
   return privKeySpec.getModulus();
  }

  public BigInteger prv_exp() {
   return privKeySpec.getPrivateExponent();
  }

  public BigInteger pub_mod() {
   return pubKeySpec.getModulus();
  }

  public BigInteger pub_exp() {
   return pubKeySpec.getPublicExponent();
  }

  public int keysize() {
   return keysize;
  }

  public static KeyPair from_file(String prvname,String pubname) {
   KeyPair res=null;

   byte prvdata[]=Utils.load_file(prvname);
   byte pubdata[]=Utils.load_file(pubname);

   if ((prvdata!=null)&&(pubdata!=null)&&(prvdata.length==pubdata.length)) {
    res=new KeyPair(prvdata,pubdata);
   }

   return res;
  }

  //verify hether a given keypair is legit for RSA crypto
  public boolean verify() {
   byte plain[]=new byte[keysize];

   for(int i=0;i<plain.length;i++) {
    plain[i]=(byte)i;
   }

   byte cipher[]=encrypt(plain,this);

   byte result_plain[]=decrypt(cipher,this);

   boolean res=true;

   for(int i=0;i<plain.length;i++) {
    if (plain[i]!=result_plain[i]) {
     res=false;
     break;
    }
   }

   return res;
  }

  public void print(String s) {
   PaddedPrinter pp=Shell.get_pp();

   pp.println(s);
   pp.pad(2,"");
   pp.println("PRV");
   pp.println("  mod: "+prv_mod().toString(16));
   pp.println("  exp: "+prv_exp().toString(16));
   pp.leave();

   pp.pad(2,"");
   pp.println("PUB");
   pp.println("  mod: "+pub_mod().toString(16));
   pp.println("  exp: "+pub_exp().toString(16));
   pp.leave();
  }
 }

 public static byte[] decrypt(byte data[],KeyPair keypair) {
  byte res[]=null;

  try {
   int dec_block_size=keypair.keysize();

   byte block[]=new byte[dec_block_size];
   byte decrypted[]=new byte[data.length];

   int size=0;

   for(int i=0;i<data.length;i+=dec_block_size) {
    System.arraycopy(data,i,block,0,dec_block_size);

    Cipher cipher=Cipher.getInstance("RSA/ECB/NoPadding");
    cipher.init(Cipher.DECRYPT_MODE,keypair.prv());
 
    byte[] plain=cipher.doFinal(block);

    System.arraycopy(plain,0,decrypted,size,plain.length);
    size+=plain.length;
   }

   res=new byte[size];
   System.arraycopy(decrypted,0,res,0,size);
  } catch(Throwable t) {t.printStackTrace();}

  return res;
 }

 public static byte[] encrypt(byte data[],KeyPair keypair) {
  byte res[]=null;

  try {
   int enc_block_size=keypair.keysize();

   byte block[]=new byte[enc_block_size];
   byte encrypted[]=new byte[2*data.length];

   int size=0;

   for(int i=0;i<data.length;i+=enc_block_size) {
    int len=enc_block_size;

    if ((data.length-i)<enc_block_size) {
     len=data.length-i;
     block=new byte[len];
    }

    System.arraycopy(data,i,block,0,len);

    Cipher cipher=Cipher.getInstance("RSA/ECB/NoPadding");
    cipher.init(Cipher.ENCRYPT_MODE,keypair.pub());
 
    byte[] ciphered=cipher.doFinal(block);

    System.arraycopy(ciphered,0,encrypted,size,ciphered.length);
    size+=ciphered.length;
   }

   res=new byte[size];
   System.arraycopy(encrypted,0,res,0,size);
  } catch(Throwable t) {t.printStackTrace();}

  return res;
 }

 public static byte[] decrypt_oaep(byte data[],KeyPair keypair) {
  byte res[]=null;

  try {
   Cipher cipher=Cipher.getInstance("RSA/ECB/OAEPWithSHA-1AndMGF1Padding");

   cipher.init(Cipher.DECRYPT_MODE,keypair.prv());
   res=cipher.doFinal(data);
  } catch(Throwable t) {t.printStackTrace();}

  return res;
 }
}
