/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package mod.cdn;

import agsecres.tool.*;
import agsecres.helper.*;
import java.lang.*;

public class Init implements TolkitMod {
 public static final CmdDesc cmd_table[]={
  new CmdDesc("urlcheck"     , "S"           ,  "url"                                               , new CMD_URLCHECK()),
  new CmdDesc("urlfetch"     , "SS"          ,  "url outputfile"                                    , new CMD_URLFETCH()),
 };

 public String name() {
  return "cdn";
 }

 public String desc() {
  return "CDN helper";
 }

 public CmdDesc[] cmd_table() {
  return cmd_table;
 }

 public static class CMD_URLCHECK implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length!=1) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   String s_url=options[0].str_arg();

   if ((s_url==null)||(s_url.equals(""))||(!s_url.startsWith("http://"))) {
    Shell.err_string="invalid url: "+s_url;
    return false;
   }

   String serial=Vars.get_str("SERIAL");

   if (serial==null) {
    Shell.err_string="SERIAL variable is empty";
    return false;
   }

   Web.PathInfo pathinfo=CDN.get_pathinfo(serial,s_url);

   if (pathinfo!=null) {
    pathinfo.print();
   } else {
    Shell.err_string="cannot get url info ["+s_url+"]";
    return false;
   }
   
   return true;
  }
 }

 public static class CMD_URLFETCH implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length!=2) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   String s_url=options[0].str_arg();
   String outputfile=options[1].str_arg();

   if ((s_url==null)||(s_url.equals(""))||(!s_url.startsWith("http://"))) {
    Shell.err_string="invalid url: "+s_url;
    return false;
   }

   String serial=Vars.get_str("SERIAL");

   if (serial==null) {
    Shell.err_string="SERIAL variable is empty";
    return false;
   }

   long size=CDN.download_content(serial,s_url,outputfile);

   if (size<0) {
    Shell.err_string="download failed";
    return false;
   }

   return true;
  }
 }
}