/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package mod.cgaweb;

import agsecres.tool.*;
import agsecres.helper.*;
import java.lang.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;

public class CGAWeb {
 public static class Perm {
  String numcol;
  String allow;

  public Perm(String numcol,String allow) {
   this.numcol=numcol;
   this.allow=allow;
  }

  public String numcol() {
   return numcol;
  }

  public String allow() {
   return allow;
  }
 }

 public static class CollectionPerms {
  String sc;
  Vector<Perm> perms;
  Hashtable<String,String> allowed;

  public CollectionPerms(String sc,Vector<Perm> perms) {
   this.sc=sc;
   this.perms=perms;

   allowed=new Hashtable<String,String>();

   for(int i=0;i<perms.size();i++) {
    Perm perm=perms.elementAt(i);

    if (perm.allow().equals("true")) {
     allowed.put(perm.numcol(),perm.allow());
    }
   }
  }

  public String scard() {
   return sc;
  }

  public boolean allowed(String colid) {
   if (colid!=null) {
    if (allowed.get(colid)!=null) return true;
   }

   return false;
  }

  public void print() {
   PaddedPrinter pp=Shell.get_pp();

   pp.println("Permissions");
   pp.pad(2,"");

   for(int i=0;i<perms.size();i++) {
    Perm perm=perms.elementAt(i);

    String allow="ALLOW";
    if (perm.allow().equals("false")) allow="NOT ALLOW";

    pp.println("col "+Utils.padded_string(perm.numcol(),4)+"  "+allow);
   }
  }
 }

 public static final String AUTHENTICATE_URL = "https://nvs1.ncplus.pl/wsi/api/v2/smartcard/authenticate";
 public static final String CGAWEB_URL       = "https://dek.ncplus.pl:8082/CGAWebOrderService/CGAWebOrderInterface";

 public static String[] get_json_reqprops() {
  return new String[]{
   "Content-type",      "application/json",
  };
 }

 public static String[] get_json_api_reqprops() {
  return new String[]{
   "Content-type",      "application/json",
   "WSI-Application",   "NCPGO20TV"
  };
 }
 public static String[] get_reqprops() {
  return new String[]{
   "Content-type",  "text/xml; charset=utf-8"
  };
 }

 public static byte[] do_get_json(String url) {
  return Web.https_get(url,get_json_reqprops());
 }

 public static String do_post_json_api(String url,String msg) {
  return Web.https_post(url,msg,get_json_api_reqprops());
 }

 public static String do_post(String url,String msg) {
  return Web.https_post(url,msg,get_reqprops());
 }

 public static String get_accountid(String sc) {
  String id=null;

  String s_url=AUTHENTICATE_URL;

  try {
   String xml="";

   xml+="{\"smartCardCode\":\""+sc+"\",\"portalClientId\":\"NCPGO20\"}";

   String res=do_post_json_api(s_url,xml);

   JSON.JSONObject root=JSON.parse(res);

   id=JSON.get_string(root,"id");
  } catch(Throwable t) {t.printStackTrace();}

  return id;
 }

 public static CollectionPerms get_perms(String sc) {
  CollectionPerms cperms=null;

  String s_url=CGAWEB_URL;

  try {
   String xml="";

   xml+="<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n";
   xml+="<soapenv:Envelope xmlns:soapenv='http://schemas.xmlsoap.org/soap/envelope/'";
   xml+="                xmlns:vnfs='http://www.ncplus.pl/CGAWebOrderInterface/'>";
   xml+="<soapenv:Header/>";
   xml+="<soapenv:Body>";
   xml+="<vnfs:getCol>";
   xml+="<numdec>"+sc+"</numdec>";
   xml+="</vnfs:getCol>";
   xml+="</soapenv:Body>";
   xml+="</soapenv:Envelope>";

   String res=do_post(s_url,xml);

   //byte data[]=Utils.load_file("xml.txt");

   Document doc=XmlUtils.parse_xml(new ByteArrayInputStream(res.getBytes()));

   Node node[]=XmlUtils.select(doc,"S:Envelope.S:Body.ns2:getColResponse.return.collst.col");

   if ((node!=null)&&(node.length>0)) {
    Vector<Perm> perms=new Vector<Perm>();

    for(int i=0;i<node.length;i++) {
     Node perm_node=node[i];

     String numcol=XmlUtils.get_value(perm_node,"numcol");
     String allow=XmlUtils.get_value(perm_node,"allow");

     Perm perm=new Perm(numcol,allow);
     perms.add(perm);
    }

    cperms=new CollectionPerms(sc,perms);
   }
  } catch(Throwable t) {t.printStackTrace();}

  return cperms;
 }

 public static boolean show_perms(String sc) {
  CollectionPerms cperms=get_perms(sc);

  if (cperms==null) {
   Shell.err_string="cannot retrieve permssions";
   return false;   
  }

  cperms.print();

  return true;
 }

 public static boolean show_tvod(String sc) throws Throwable {
  Shell.err_string="REMOVE FROM THIS RELEASE";
  return false;   
 }
}
