/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package mod.cgaweb;

import agsecres.tool.*;
import mod.cdn.*;
import java.lang.*;
import java.math.*;

public class Init implements TolkitMod {
 public static final CmdDesc cmd_table[]={
  new CmdDesc("accountid"      , "sS"           ,  "[-s scard]"                                                         , new CMD_ACCOUNTID()),
  new CmdDesc("cgaperms"       , "sS"           ,  "[-s scard]"                                                         , new CMD_CGAPERMS()),
  new CmdDesc("tvodinfo"       , "sS"           ,  "[-s scard]"                                                         , new CMD_TVODINFO()),
 };

 public String name() {
  return "cgaweb";
 }

 public String desc() {
  return "CANALP CGAWeb toolkit";
 }

 public CmdDesc[] cmd_table() {
  return cmd_table;
 }

 public static String cur_scard() {
  return Vars.get_str("SCARD");
 }

 private static String parse_scard(Shell.Option[] options) {
  String scard=null;

  Shell.Option opt=Shell.getopt(options,'s');

  if (opt!=null) {
   scard=opt.str_arg();
  } else {
   scard=cur_scard();
  }

  if (scard==null) {
   Shell.err_string="missing scard argument";
   return null;
  }

  return scard;
 }

 public static class CMD_ACCOUNTID implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String scard=parse_scard(options);

   String accountid=CGAWeb.get_accountid(scard);

   if (accountid!=null) {
    Shell.println("scard:      "+scard);
    Shell.println("accountid:  "+accountid);
   }

   return true;
  }
 }

 public static class CMD_CGAPERMS implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String scard=parse_scard(options);

   return CGAWeb.show_perms(scard);
  }
 }

 public static class CMD_TVODINFO implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String scard=parse_scard(options);

   return CGAWeb.show_tvod(scard);
  }
 }
}