/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package mod.mspr;

import agsecres.tool.*;
import java.lang.*;

public class Codec {
 public static final int AACL_CODEC = 0x4141434C;
 public static final int AVC1_CODEC = 0x41564331;

 public static abstract class Audio extends Codec {
  short channel_count;
  short sample_size;
  int sample_rate;

  public void set_channel_count(short channel_count) {
   this.channel_count=channel_count;
  }

  public void set_sample_size(short sample_size) {
   this.sample_size=sample_size;
  }

  public void set_sample_rate(int sample_rate) {
   this.sample_rate=sample_rate<<16;
  }

  public static Audio get(int type,short channel_count,short sample_size,int sample_rate) {
   Codec codec=Codec.get(type);

   if ((codec!=null)&&(codec instanceof Audio)) {
    Audio audio=(Audio)codec;

    audio.set_channel_count(channel_count);
    audio.set_sample_size(sample_size);
    audio.set_sample_rate(sample_rate);

    return audio;
   }

   return null;
  }

  public MP4File.SampleEntry get_sample_entry(byte codec_prv_data[]) {
   //MPEG2 AAC-Low Complexity Profile
   byte objectType=0x67;

   byte streamType=0x15;
   byte upStream=0;

   int bufferSizeDB=0x00000600;
   int maxBitrate=0x00177000;
   int avgBitrate=0x00177000;

   byte slconfigdata[]=new byte[]{
    0x02
   };

   MP4File.DecoderSpecificInfo dsinfo=new MP4File.DecoderSpecificInfo(codec_prv_data);

   MP4File.DecoderConfigDescr dcdesc=new MP4File.DecoderConfigDescr(objectType,streamType,upStream,bufferSizeDB,maxBitrate,avgBitrate,dsinfo);
   MP4File.SLConfigDescriptor slcdesc=new MP4File.SLConfigDescriptor(slconfigdata);

   short es_id=0;
   byte flags=0;
   byte priority=0;
 
   MP4File.ESDescriptor esdesc=new MP4File.ESDescriptor(es_id,flags,priority,dcdesc,slcdesc);

   MP4File.ESDS esds=new MP4File.ESDS(esdesc);

   short data_reference_index=1;
   MP4File.MP4A mp4a=new MP4File.MP4A(data_reference_index,channel_count,sample_size,sample_rate,esds);
   MP4File.AudioSampleEntry ase=new MP4File.AudioSampleEntry(mp4a);

   return ase;
  }
 }

 public static abstract class Video extends Codec {
  public static final byte CODEC_PRV_DATA_PATTERN[]={
   0x00, 0x00, 0x00, 0x01
  };

  short width;
  short height;

  public void set_width(short width) {
   this.width=width;
  }

  public void set_height(short height) {
   this.height=height;
  }

  public static Video get(int type,short width,short height) {
   Codec codec=Codec.get(type);

   if ((codec!=null)&&(codec instanceof Video)) {
    Video video=(Video)codec;

    video.set_width(width);
    video.set_height(height);

    return video;
   }

   return null;
  }

  private static boolean pattern_match(byte data[],byte pattern[],int off) {
   if ((off+pattern.length)>data.length) return false;

   for(int i=0;i<pattern.length;i++) {
    if (data[off+i]!=pattern[i]) return false;
   }

   return true;
  }

  private static int find_pattern(byte data[],byte pattern[],int off) {
   for(int pos=0;pos<(data.length-pattern.length);pos++) {
    if (pattern_match(data,pattern,off+pos)) return (off+pos+pattern.length);
   }

   return -1;
  }

  public MP4File.SampleEntry get_sample_entry(byte codec_prv_data[]) {
   int sps_off=find_pattern(codec_prv_data,CODEC_PRV_DATA_PATTERN,0);
   int pps_off=find_pattern(codec_prv_data,CODEC_PRV_DATA_PATTERN,sps_off);

   int sps_len=pps_off-sps_off-CODEC_PRV_DATA_PATTERN.length;
   int pps_len=codec_prv_data.length-pps_off;

   if ((sps_off>=0)&&(pps_off>=0)&&(sps_len>=0)&&(pps_len>=0)) {
    byte sequenceParameterSetNALUnit[]=new byte[sps_len];
    System.arraycopy(codec_prv_data,sps_off,sequenceParameterSetNALUnit,0,sps_len);

    byte pictureParameterSetNALUnit[]=new byte[pps_len];
    System.arraycopy(codec_prv_data,pps_off,pictureParameterSetNALUnit,0,pps_len);
 
    byte config_version=1;
    byte AVCProfileIndication=0x64;
    byte profile_compatibility=0x40;
    byte AVCLevelIndication=0x0d;
    byte lengthSizeMinusOne=3;

    MP4File.AVCC avcc=new MP4File.AVCC(config_version,AVCProfileIndication,profile_compatibility,AVCLevelIndication,
                                       lengthSizeMinusOne,sequenceParameterSetNALUnit,pictureParameterSetNALUnit);

    short data_reference_index=1;
    MP4File.AVC1 avc1=new MP4File.AVC1(data_reference_index,width,height,avcc);

    MP4File.VisualSampleEntry vse=new MP4File.VisualSampleEntry(avc1);

    return vse;    
   }

   return null;
  }
 }

 public static class AACLCodec extends Audio {
  public void decrypt(MP4File mp4,byte content_key[]) throws Throwable {
   if (!verify_sample_cnt(mp4)) ERR.log("Inconsistent sample_cnt in TRUN and SEB boxes");
   if (!verify_mdat(mp4)) ERR.log("Inconsistent sample info sizes in TRUN and MDAT length");

   MP4File.TRUN trun=mp4.get_trun();
   MP4File.SampleEncryptionBox seb=mp4.get_seb();
   MP4File.MDAT mdat=mp4.get_mdat();

   byte data[]=mdat.data();

   int cnt=trun.sample_cnt();

   int off=0;

   for(int idx=0;idx<cnt;idx++) {
    MP4File.SampleInfo si=trun.get_sinfo(idx);
    MP4File.IV iv=seb.get_iv(idx);

    int size=si.size();

    byte[] plaintext=MSPR.aes_ctr_decrypt(data,off,size,iv.data(),content_key);
    System.arraycopy(plaintext,0,data,off,size);

    off+=size;
   }
  }
 }

 public static class AVC1Codec extends Video {
  public boolean verify_iv2(MP4File mp4) {
   MP4File.MDAT mdat=mp4.get_mdat();
   MP4File.SampleEncryptionBox seb=mp4.get_seb();

   if ((mdat!=null)&&(seb!=null)) {
    int total_len=0;

    for(int i=0;i<seb.iv_cnt();i++) {
     MP4File.IV_2 iv2=(MP4File.IV_2)seb.get_iv(i);

     for(int j=0;j<iv2.slice_cnt();j++) {
      MP4File.SliceInfo si=iv2.get_slice(j);

      total_len+=si.clear_data_len();
      total_len+=si.encrypted_data_len();
     }
    }

    int mdat_len=mdat.data().length;

    return total_len==mdat_len;
   }

   return false;
  }

  public void decrypt(MP4File mp4,byte content_key[]) throws Throwable {
   if (!verify_sample_cnt(mp4)) ERR.log("Inconsistent sample_cnt in TRUN and SEB boxes");
   if (!verify_mdat(mp4)) ERR.log("Inconsistent sample info sizes in TRUN and MDAT length");
   if (!verify_iv2(mp4)) ERR.log("Inconsistent data length in SEB and MDAT boxes");

   MP4File.SampleEncryptionBox seb=mp4.get_seb();
   MP4File.MDAT mdat=mp4.get_mdat();

   byte data[]=mdat.data();

   int off=0;

   for(int i=0;i<seb.iv_cnt();i++) {
    MP4File.IV_2 iv2=(MP4File.IV_2)seb.get_iv(i);

    int slice_len=iv2.slice_len();

    //prepare slice data for decryption
    byte ciphertext[]=new byte[slice_len];

    int src_off=off;
    int dst_off=0;

    for(int j=0;j<iv2.slice_cnt();j++) {
     MP4File.SliceInfo si=iv2.get_slice(j);

     src_off+=si.clear_data_len();

     int size=si.encrypted_data_len();

     System.arraycopy(data,src_off,ciphertext,dst_off,size);

     src_off+=size;
     dst_off+=size;
    }

    //decrypt slice
    byte[] plaintext=MSPR.aes_ctr_decrypt(ciphertext,0,slice_len,iv2.data(),content_key);

    //copy plaintext to proper slice locations
    src_off=0;
    dst_off=off;

    for(int j=0;j<iv2.slice_cnt();j++) {
     MP4File.SliceInfo si=iv2.get_slice(j);

     dst_off+=si.clear_data_len();

     int size=si.encrypted_data_len();

     System.arraycopy(plaintext,src_off,data,dst_off,size);

     src_off+=size;
     dst_off+=size;
    }

    off+=slice_len;
   }
  }
 }

 public boolean verify_sample_cnt(MP4File mp4) {
  MP4File.TRUN trun=mp4.get_trun();
  MP4File.SampleEncryptionBox seb=mp4.get_seb();

  if ((trun!=null)&&(seb!=null)) {
   return trun.sample_cnt()==seb.sample_cnt();
  }

  return false;
 }

 public boolean verify_mdat(MP4File mp4) {
  MP4File.TRUN trun=mp4.get_trun();
  MP4File.MDAT mdat=mp4.get_mdat();

  if ((trun!=null)&&(mdat!=null)) {
   int total_len=0;

   for(int i=0;i<trun.sample_cnt();i++) {
    MP4File.SampleInfo si=trun.get_sinfo(i);

    total_len+=si.size();
   }

   int mdat_len=mdat.data().length;

   return total_len==mdat_len;
  }

  return false;
 }

 public static Codec get(int type) {
  switch(type) {
   case AACL_CODEC:
    return new AACLCodec();
   case AVC1_CODEC:
    return new AVC1Codec();
  }

  ERR.log("Unsupported codec type: "+Utils.hex_value(type,8));

  return null;
 }

 public void decrypt(MP4File mp4,byte content_key[]) throws Throwable {
  ERR.log("Unimplemented Codec.decrypt");
 }
}
