/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package mod.mspr;

import agsecres.tool.*;
import agsecres.helper.*;
import java.lang.*;
import java.io.*;

public class KeyExtractor {
 public static class KeyConfig {
  String keyfile;
  String start_sym;
  String end_sym;

  public KeyConfig(String keyfile,String start_sym,String end_sym) {
   this.keyfile=keyfile;
   this.start_sym=start_sym;
   this.end_sym=end_sym;
  }
 }

 static KeyConfig K1_CONFIG = new KeyConfig("k1", "__img_keys_start",     "__img_template_start");
 static KeyConfig T1_CONFIG = new KeyConfig("t1", "__img_template_start", "__img_template_end");
 static KeyConfig G1_CONFIG = new KeyConfig("g1", "__img_template_end",   "__img_bgc_end");
 static KeyConfig Z1_CONFIG = new KeyConfig("z1", "__img_bgc_end",        "__img_zgp_end");

 public static byte[] extract_data(ELF elf,KeyConfig kc) {
  ELF.Symbol sym1=elf.lookup_symbol(kc.start_sym);
  ELF.Symbol sym2=elf.lookup_symbol(kc.end_sym);

  if ((sym1!=null)&&(sym2!=null)) {
   int off=sym1.value();
   int size=sym2.value()-off;

   return elf.read_n(off,size);
  }
   
  return null;
 }

 public static boolean extract_keys(String binarypath,RSA.KeyPair keypair) {
  try {
   ELF elf=ELF.load(binarypath);

   PaddedPrinter pp=PaddedPrinter.getInstance();

   byte k1_enc_data[]=extract_data(elf,K1_CONFIG);

   if (k1_enc_data==null) {
    Shell.err_string="cannot find K1 encrypted data";
    return false;
   }

   pp.println("- decrypting keys image (K1)");

   byte k1_data[]=RSA.decrypt_oaep(k1_enc_data,keypair);

   if (k1_data==null) {
    Shell.err_string="failed to decrypt K1 encrypted data";
    return false;
   }

   pp.printhex("K1",k1_data);

   byte key[]=new byte [0x10];
   byte iv[]=new byte [0x10];

   System.arraycopy(k1_data,0x00,key,0,0x10);
   System.arraycopy(k1_data,0x50,iv,0,0x10);

   pp.printhex("aes key",key);
   pp.printhex("aes iv",iv);

   byte t1_enc_data[]=extract_data(elf,T1_CONFIG);

   if (t1_enc_data==null) {
    Shell.err_string="cannot find T1 encrypted data";
    return false;
   }

   pp.println("- decrypting template certificate (T1)");

   byte t1_data[]=Crypto.aes_cbc_decrypt(t1_enc_data,iv,key);

   if (t1_data==null) {
    Shell.err_string="failed to decrypt T1 encrypted data";
    return false;
   }

   pp.printhex("T1",t1_data);

   byte g1_enc_data[]=extract_data(elf,G1_CONFIG);

   if (g1_enc_data==null) {
    Shell.err_string="cannot find G1 encrypted data";
    return false;
   }

   pp.println("- decrypting binary group certificate (G1)");

   byte g1_data[]=Crypto.aes_cbc_decrypt(g1_enc_data,iv,key);

   if (g1_data==null) {
    Shell.err_string="failed to decrypt G1 encrypted data";
    return false;
   }

   pp.printhex("G1",g1_data);

   byte z1_enc_data[]=extract_data(elf,Z1_CONFIG);

   if (z1_enc_data==null) {
    Shell.err_string="cannot find Z1 encrypted data";
    return false;
   }

   pp.println("- decrypting private ECC group key (Z1)");

   byte z1_data[]=Crypto.aes_cbc_decrypt(z1_enc_data,iv,key);

   if (z1_data==null) {
    Shell.err_string="failed to decrypt Z1 encrypted data";
    return false;
   }

   pp.printhex("Z1",z1_data);
  } catch(Throwable t) {
  }

  return true;
 }
}
