/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package mod.vod;

import agsecres.tool.*;
import mod.cdn.*;
import java.lang.*;
import java.util.*;
import java.math.*;

public class Init implements TolkitMod {
 public static final CmdDesc cmd_table[]={
  new CmdDesc("vodinfo"     , null           ,  null                                                         , new CMD_VODINFO()),
  new CmdDesc("colinfo"     , "S"            ,  "coluniqueid"                                                , new CMD_COLINFO()),
  new CmdDesc("lscol"       , "SI"           ,  "coluniqueid [pagecnt]"                                      , new CMD_LSCOL()),
  new CmdDesc("lspage"      , "SI"           ,  "coluniqueid pageid"                                         , new CMD_LSPAGE()),
  new CmdDesc("colasset"    , "S"            ,  "assetid"                                                    , new CMD_COLASSET()),
  new CmdDesc("storeasset"  , "S"            ,  "assetid"                                                    , new CMD_STOREASSET()),
 };

 public String name() {
  return "vod";
 }

 public String desc() {
  return "CANALP VOD toolkit";
 }

 public CmdDesc[] cmd_table() {
  return cmd_table;
 }

 public static class CMD_VODINFO implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   VOD.Collection root=VOD.vod_root();

   root.print(false);

   return true;
  }
 }

 public static class CMD_COLINFO implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length!=1) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   String uniqueid=options[0].str_arg();

   if ((uniqueid==null)||(!uniqueid.startsWith("COL"))) {
    Shell.err_string="invalid collection id: "+uniqueid;
    return false;
   }

   VOD.Collection root=VOD.vod_root();

   VOD.Collection col=VOD.get_collection(root,uniqueid);

   if (col==null) {
    Shell.err_string="collection cannot be found: "+uniqueid;
    return false;
   }

   col.print(true);

   return true;
  }
 }

 public static class CMD_LSCOL implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if ((options.length!=1)&&(options.length!=2)) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   int page_cnt=0;

   String uniqueid=options[0].str_arg();

   if ((uniqueid==null)||(!uniqueid.startsWith("COL"))) {
    Shell.err_string="invalid collection id: "+uniqueid;
    return false;
   }

   if (options.length==2) {
    page_cnt=options[1].int_arg();
   }

   VOD.Collection root=VOD.vod_root();

   VOD.Collection col=VOD.get_collection(root,uniqueid);

   if (col==null) {
    Shell.err_string="collection cannot be found: "+uniqueid;
    return false;
   }

   Vector<VOD.Asset> assets=col.assets(page_cnt);

   PaddedPrinter pp=Shell.get_pp();

   String hdr="ASSETS from "+col.page_num()+"/"+col.max_page()+" pages";

   pp.println(hdr);
   pp.pad(2,"");

   for(int i=0;i<assets.size();i++) {
    VOD.Asset asset=assets.elementAt(i);

    String line=Utils.padded_string(asset.id(),16)+" \""+asset.title()+"\"";

    pp.println(line);
   }

   pp.leave();

   return true;
  }
 }

 public static class CMD_LSPAGE implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length!=2) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   int page_id=1;

   String uniqueid=options[0].str_arg();

   if ((uniqueid==null)||(!uniqueid.startsWith("COL"))) {
    Shell.err_string="invalid collection id: "+uniqueid;
    return false;
   }

   page_id=options[1].int_arg();

   VOD.Collection root=VOD.vod_root();

   VOD.Collection col=VOD.get_collection(root,uniqueid);

   if (col==null) {
    Shell.err_string="collection cannot be found: "+uniqueid;
    return false;
   }

   Vector<VOD.Asset> assets=col.assets_from_page(page_id);

   PaddedPrinter pp=Shell.get_pp();

   String hdr="ASSETS from "+page_id+"/"+col.max_page()+" pages";

   pp.println(hdr);
   pp.pad(2,"");

   for(int i=0;i<assets.size();i++) {
    VOD.Asset asset=assets.elementAt(i);

    String line=Utils.padded_string(asset.id(),16)+" \""+asset.title()+"\"";

    pp.println(line);
   }

   pp.leave();

   return true;
  }
 }

 public static class CMD_COLASSET implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length!=1) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   String assetid=options[0].str_arg();

   if (assetid==null) {
    Shell.err_string="invalid asset id: "+assetid;
    return false;
   }

   VOD.Collection root=VOD.vod_root();

   VOD.Asset asset=VOD.get_asset(root,assetid);

   if (asset==null) {
    Shell.err_string="asset not yet present in DB: "+assetid;
    return false;
   }

   asset.print(true);

   return true;
  }
 }

 public static class CMD_STOREASSET implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length!=1) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   String assetid=options[0].str_arg();

   if (assetid==null) {
    Shell.err_string="invalid asset id: "+assetid;
    return false;
   }

   VOD.Collection root=VOD.vod_root();

   VOD.Asset asset=VOD.get_asset(root,assetid);

   if (asset==null) {
    Shell.err_string="asset not yet present in DB: "+assetid;
    return false;
   }

   String store_path=asset.store();

   if (store_path==null) {
    Shell.err_string="error saving asset data: "+assetid;
    return false;    
   }

   Shell.println("asset data stored to: "+store_path);

   return true;
  }
 }

}