/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package mod.vod;

import agsecres.tool.*;
import agsecres.helper.*;
import mod.cgaweb.*;
import java.lang.*;
import java.io.*;
import java.util.*;

public class VOD {
 public static final String PAGE_TYPE = "ADB_page";
 public static final String SORT_TYPE = "alphabet";

 public static class AssetBrowseInfo {
  String browse_url;
  int max_page;

  public AssetBrowseInfo(String browse_url,int max_page) {
   this.browse_url=browse_url;
   this.max_page=max_page;
  }

  public int max_page() {
   return max_page;
  }

  public String page_url(int page_id) {
   return browse_url+page_id+"%26ppp%3D30";
  }
 }

 public static class Collection {
  String attname;
  String attid;
  String attlink;
  String attproxy;
  String atttVODCol;

  String uniqueid;

  String count;

  String browseurl;

  Collection parent;
  Vector<Collection> children;

  String page_num;
  String max_page;

  Vector<Asset> assets;

  public Collection() {
   children=new Vector<Collection>();
  }

  public Collection(String name) {
   this();

   this.attname=name;
  }

  public Collection(String name,Vector<Collection> children) {
   this.attname=name;
 
   this.children=children;
  }

  public void set_children(Vector<Collection> children) {
   this.children=children;
  }

  public int subcol_cnt() {
   return children.size();
  }

  public Collection get_subcol(int i) {
   if (i<subcol_cnt()) {
    return children.elementAt(i);
   }
 
   return null;
  }
  
  public String id() {
   return attid;
  }

  public String name() {
   return attname;
  }

  public String link() {
   return attlink;
  }

  public String proxy() {
   return attproxy;
  }

  public String count() {
   if (count==null) {
    fetch_category_attributes();
   }
    
   return count;
  }

  public String browseurl() {
   if (browseurl==null) {
    fetch_category_attributes();
   }

   return browseurl;
  }

  public String uniqueid() {
   return uniqueid;
  }

  public String page_num() {
   return page_num;
  }

  public String max_page() {
   return max_page;
  }

  public Collection parent() {
   return parent;
  }

  public void set_parent(Collection parent) {
   this.parent=parent;
  }

  public void set_count(String count) {
   this.count=count;
  }

  public void set_browseurl(String browseurl) {
   this.browseurl=browseurl;
  }

  public void set_uniqueid() {
   uniqueid=next_id();
  }

  //set count and browse url
  public void fetch_category_attributes() {
   try {
    if ((uniqueid()!=null)&&(link()!=null)) {
     String url=link();

     byte xml[]=do_get_json(url);

     if (xml!=null) {
      JSON.JSONObject root=JSON.parse(xml);

      String type=JSON.get_string(root,"type");

      if (type.equals("categories")) {
       JSON.JSONBucket category=JSON.get_bucket(root,"category");

       if (category!=null) {
        JSON.JSONArray pages=JSON.get_array(category,PAGE_TYPE);
        String count=JSON.get_string(category,"count");

        set_count(count);
  
        JSON.JSONObject browsing_page=null;

        for(int i=0;i<pages.size();i++) {
         JSON.JSONObject page=pages.get(i);
 
         String sort=JSON.get_string(page,"sort");

         if (sort.equals(SORT_TYPE)) {
          browsing_page=page;
          break;
         }
        }

        if (browsing_page!=null) {
         String browse_url=JSON.get_string(browsing_page,"value");

         if (browse_url!=null) {
          set_browseurl(browse_url);
         }
        }
       }
      }
     }
    }
   } catch(Throwable t) {
    t.printStackTrace();
   }
  }

  public void fetch_pages(String url,int page_cnt) {
   try {
    int num_pages=0;

    while(true) {
     byte xml[]=do_get_json(url);

     JSON.JSONObject root=JSON.parse(xml);

     String type=JSON.get_string(root,"type");

     if (type.equals("page")) {
      num_pages++;

      JSON.JSONBucket page=JSON.get_bucket(root,"page");

      page_num=JSON.get_string(page,"number");
      max_page=JSON.get_string(page,"max_page");

      String next=JSON.get_string(page,"next");

      JSON.JSONArray contents=JSON.get_array(root,"contents");

      for(int i=0;i<contents.size();i++) {
       JSON.JSONObject content=contents.get(i);

       String id=JSON.get_string(content,"id");
       String title=JSON.get_string(content,"title");
       String link=JSON.get_string(content,"link");

       Asset asset=new Asset(id,title,link,proxy());
       add_asset(asset);
      }

      if (page_cnt!=0) {
       if (num_pages>=page_cnt) break;
      }

      if (next!=null) {
       url=next;
      } else break;

     } else break;
    }
   } catch(Throwable t) {
    t.printStackTrace();
   }
  }

  public AssetBrowseInfo get_asset_browse_info() {
   try {
    String url=browseurl();

    if (url!=null) {
     byte xml[]=do_get_json(url);

     JSON.JSONObject root=JSON.parse(xml);

     String type=JSON.get_string(root,"type");

     if (type.equals("page")) {
      JSON.JSONBucket page=JSON.get_bucket(root,"page");

      String max_page=JSON.get_string(page,"max_page");
      String next=JSON.get_string(page,"next");

      int page_cnt=Integer.parseInt(max_page);
      int idx=next.indexOf("%26page%3D");

      if (idx>0) {
       String base_url=next.substring(0,idx+10);
       return new AssetBrowseInfo(base_url,page_cnt);
      }
     }
    }
   } catch(Throwable t) {
   }

   return null;
  }

  public void fetch_assets_from_page(int page_id) {
   try {
    AssetBrowseInfo abi=get_asset_browse_info();

    if (abi!=null) {
     fetch_pages(abi.page_url(page_id),1);
    }
   } catch(Throwable t) {
    t.printStackTrace();
   }
  }

  public void fetch_assets(int page_cnt) {
   try {
    String url=browseurl();

    if (url!=null) {
     fetch_pages(url,page_cnt);
    }
   } catch(Throwable t) {
    t.printStackTrace();
   }
  }

  public void add_asset(Asset asset) {
   assets.add(asset);
  }

  public Vector<Asset> assets(int page_cnt) {
   assets=new Vector<Asset>();

   fetch_assets(page_cnt);
 
   return assets;
  }

  public Vector<Asset> assets_from_page(int page_id) {
   assets=new Vector<Asset>();

   fetch_assets_from_page(page_id);
 
   return assets;
  }

  public void print(boolean verbose) {
   PaddedPrinter pp=Shell.get_pp();

   String hdr="";

   CGAWeb.CollectionPerms cperms=get_cperms();
   String allowed=" ";

   if (cperms.allowed(id())) allowed="*";

   if (subcol_cnt()>0) {
    hdr="["+name()+"]";
   } else {
    hdr=Utils.padded_string(uniqueid,6)+" "+allowed+" \""+name()+"\"";
   }

   pp.println(hdr);
   pp.pad(2,"");

   if (verbose) {
    if (attid!=null)     pp.println("id:         "+attid);
    if (attlink!=null)   pp.println("link:       "+attlink);
    if (attproxy!=null)  pp.println("proxy:      "+attproxy);
    if (count()!=null)     pp.println("count:      "+count);
    if (browseurl()!=null) pp.println("browse_url: "+browseurl);
   }

   if (subcol_cnt()>0) {
    for(int i=0;i<subcol_cnt();i++) {
     Collection col=get_subcol(i);
     col.print(verbose);
    }
   }

   pp.leave();
  }
 }

 public static class Asset {
  String id;
  String title;
  String link;
  String ls_url;

  public Asset(String id,String title,String link,String ls_url) {
   this.id=id;
   this.title=title;
   this.link=link;
   this.ls_url=ls_url;
  }

  public String id() {
   return id;
  }

  public String title() {
   return title;
  }

  public String link() {
   return link;
  }

  public String ls_url() {
   return ls_url;
  }

  public String get_stream_url() throws Throwable {
   String url=link();

   if (url!=null) {
    byte xml[]=do_get_json(url);

    JSON.JSONObject root=JSON.parse(xml);
 
    if (root!=null) { 
     return JSON.get_string(root,"stream_low");
    }
   }
  
   return null;
  }

  public String store() throws Throwable {
   String assetid=id();
   String ls=ls_url();

   String stream_url=null;
   byte xml[]=null;

   String url=link();

   if (url!=null) {
    xml=do_get_json(url);

    JSON.JSONObject root=JSON.parse(xml);
 
    if (root!=null) { 
     stream_url=JSON.get_string(root,"stream_low");
    }
   }

   if (stream_url==null) return null;

   String assetdir=asset_dir(assetid);
   Utils.mkdir(assetdir);

   String url_filename=url_filename(assetid);
   String lsurl_filename=lsurl_filename(assetid);
   String info_filename=info_filename(assetid);

   Utils.save_file(url_filename,stream_url.getBytes());
   Utils.save_file(lsurl_filename,ls.getBytes());

   if (xml!=null) Utils.save_file(info_filename,xml);

   return assetdir;
  }

  public void print(boolean verbose) {
   PaddedPrinter pp=Shell.get_pp();

   pp.println(id+" \""+title+"\"");

   if (verbose) {
    pp.pad(2,"");
    pp.println("link:    "+link);
    pp.println("ls_url:  "+ls_url);
   }
  
   pp.leave();
  } 
 }

 public static final String ID_PREFIX = "COL";
 public static Collection root;

 static int id=1;

 public static final String DEFAULT_CACHE_DIR = "content";

 //required content
 public static final String URL_FILE      = "url.txt";
 public static final String LS_FILE       = "ls.txt";
 public static final String INFO_FILE     = "Info.json";

 //collection permission
 static CGAWeb.CollectionPerms cperms;

 public static String content_dir() {
  String cache_dir=Vars.get_str("CONTENT_DIR");

  if (cache_dir==null) {
   cache_dir=DEFAULT_CACHE_DIR;
  }

  return cache_dir;
 }

 public static String asset_dir(String assetname) {
  String cache_dir=content_dir();

  String dir=cache_dir+File.separatorChar+assetname;

  return dir;
 }

 public static String url_filename(String assetname) {
  String file=asset_dir(assetname)+File.separatorChar+URL_FILE;

  return file;
 }

 public static String lsurl_filename(String assetname) {
  String file=asset_dir(assetname)+File.separatorChar+LS_FILE;

  return file;
 }

 public static String info_filename(String assetname) {
  String file=asset_dir(assetname)+File.separatorChar+INFO_FILE;

  return file;
 }

 public static String next_id() {
  String str_id=ID_PREFIX+Utils.hex_value(id,2);
  id++;

  return str_id;
 }

 public static String[] get_json_reqprops() {
  return new String[]{
   "Content-type",      "application/json",
  };
 }

 public static byte[] do_get_json(String url) {
  return Web.https_get(url,get_json_reqprops());
 }

 static boolean is_category_url(String url) {
  return url.contains("stb_categories.json");
 }

 public static Vector<Collection> get_collections(Collection parent,String url) throws Throwable {
  Vector<Collection> collections=new Vector<Collection>();

  try {
   byte xml[]=do_get_json(url);

   if (xml!=null) {
    JSON.JSONObject root=JSON.parse(xml);

    String type=JSON.get_string(root,"type");

    if (type.equals("collections")) {
     JSON.JSONArray json_collections=JSON.get_array(root,"collections");

     for(int i=0;i<json_collections.size();i++) {
      JSON.JSONObject col=json_collections.get(i);

      Collection collection=(Collection)JSON.instance_from_node(Collection.class,col);

      if (collection!=null) {
       collection.set_parent(parent);

       collections.add(collection);
      }
     }

     if (parent!=null) parent.set_children(collections);

     //process children
     for(int i=0;i<collections.size();i++) {
      Collection child=collections.elementAt(i);

      String child_url=child.link();

      if (!is_category_url(child_url)) {
       get_collections(child,child_url);
      } else {
       child.set_uniqueid();
      }
     }
    } else

    if (type.equals("categories")) {
    }

   }
  } catch(Throwable t) {
   t.printStackTrace();
  }

  return collections;
 }

 public static Collection get_vod_root(String name,String url) throws Throwable {
  Vector<Collection> collections=get_collections(null,url);

  Collection root=new Collection(name,collections);

  return root;
 }

 public static Collection vod_root() throws Throwable {
  if (root==null) {
   String vod_url=Vars.get_str("VOD_URL");

   if (vod_url!=null) {
    root=get_vod_root("VOD LIBRARY",vod_url);
   }
  }

  return root;
 }

 public static Collection get_collection(Collection from,String uniqueid) {
  for(int i=0;i<from.subcol_cnt();i++) {
   Collection col=from.get_subcol(i);

   if (col.uniqueid()!=null) {
    if (col.uniqueid().equals(uniqueid)) return col;
   }

   //search children
   Collection child_col=get_collection(col,uniqueid);

   if (child_col!=null) return child_col;
  }

  return null;  
 }

 public static Collection get_collection_by_name(Collection from,String name) {
  for(int i=0;i<from.subcol_cnt();i++) {
   Collection col=from.get_subcol(i);

   if (col.uniqueid()!=null) {
    if (col.name().equals(name)) return col;
   }

   //search children
   Collection child_col=get_collection_by_name(col,name);

   if (child_col!=null) return child_col;
  }

  return null;  
 }

 public static Asset get_asset(Collection from,String id) {
  for(int i=0;i<from.subcol_cnt();i++) {
   Collection col=from.get_subcol(i);

   Vector<VOD.Asset> assets=col.assets;

   if (assets!=null) {
    for(int j=0;j<assets.size();j++) {
     Asset asset=assets.elementAt(j);

     if (asset.id().equals(id)) return asset;
    }
   }

   //search children
   Asset child_asset=get_asset(col,id);

   if (child_asset!=null) return child_asset;
  }

  return null;  
 }

 public static CGAWeb.CollectionPerms get_cperms() {
  boolean sc_changed=false;

  String scard=Vars.get_str("SCARD");

  if (cperms!=null) {
   if (!cperms.scard().equals(scard)) sc_changed=true;
  }

  if ((cperms==null)||sc_changed) {
   cperms=CGAWeb.get_perms(scard);
  }

  return cperms;
 }
}